This is a TypeScript AST transformer [1] that assigns a name to each arrow/anonymous function and class exported as default (e.g. `export default () ⇒ 42`). The name is derived from the name of the source file or its directory, in the case of `index` file.

This transformer does basically the same as [babel-plugin-transform-export-default-name](https://github.com/gajus/babel-plugin-transform-export-default-name) but it’s implemented for the TypeScript compiler. See [README](https://github.com/gajus/babel-plugin-transform-export-default-name/blob/v2.0.4/README.md) of the aforesaid Babel plugin for the explanation why is this useful.

Examples
========

Arrow Function
--------------

***forty-two.ts*:.**

``` ts
export default () => 42
```

This will be transformed to:

***forty-two.js*:.**

``` ts
const fortyTwo = () => 42;
export default fortyTwo;
```

Anonymous Function Declaration
------------------------------

***forty-two.ts*:.**

``` ts
export default function () {
  return 42
}
```

This will be transformed to:

***forty-two.js*:.**

``` ts
export default function fortyTwo () {
  return 42;
}
```

Anonymous Class Declaration
---------------------------

***forty-two.ts*:.**

``` ts
export default class {
  get answer() { return 42 }
}
```

This will be transformed to:

***forty-two.js*:.**

``` ts
export default class FortyTwo {
  get answer() { return 42; }
}
```

How to Use
==========

Add ts-transformer-export-default-name package to your project as a **dev**elopment dependency and register it as a custom transformer.

Unfortunately, TypeScript itself does not currently provide any easy way to use custom transformers (see [Microsoft/TypeScript\#14419](https://github.com/Microsoft/TypeScript/issues/14419)). Fortunately, there are few solutions.

TTypescript
-----------

If you don’t use any bundler such as Rollup or webpack, [TTypescript](https://github.com/cevek/ttypescript) is the way to go. It provides wrappers `ttsc` and `ttserver` for the `tsc` and `tsserver` commands that add support for custom transformers. All you have to do is to use these wrappers instead of the original commands and define the transformer in your `tsconfig.json`:

**tsconfig.json:.**

``` jsonc
{
  "compilerOptions": {
    // ...
    "plugins": [
      { "transform": "ts-transformer-export-default-name" }
    ]
  },
  // ...
}
```

Rollup (with rollup-plugin-typescript2)
---------------------------------------

**rollup.config.js:.**

``` js
import typescript from 'rollup-plugin-typescript2'
import exportDefaultNameTransformer from 'ts-transformer-export-default-name'

export default {
  // ...
  plugins: [
    typescript({
      transformers: [
        (service) => ({
          before: [ exportDefaultNameTransformer(service.getProgram()) ],
          after: [],
        }),
      ],
    }),
  ],
}
```

Webpack (with ts-loader or awesome-typescript-loader)
-----------------------------------------------------

**webpack.config.js:.**

``` js
const exportDefaultNameTransformer = require('ts-transformer-export-default-name').default

module.exports = {
  // ...
  module: {
    rules: [
      {
        test: /\.ts$/,
        loader: 'ts-loader', // or 'awesome-typescript-loader',
        options: {
          getCustomTransformers: (program) => ({
            before: [
              exportDefaultNameTransformer(program),
            ],
          }),
        },
      },
    ],
  },
}
```

Credits
=======

-   This project is highly inspired by [babel-plugin-transform-export-default-name](https://github.com/gajus/babel-plugin-transform-export-default-name).

License
=======

This project is licensed under [MIT License](http://opensource.org/licenses/MIT/). For the full text of the license, see the [LICENSE](LICENSE) file.

[1] If you’ve never heard about TypeScript transformers, I can recommend [this blog post](https://blog.logrocket.com/using-typescript-transforms-to-enrich-runtime-code-3fd2863221ed/) to dive into the topic.
