"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs = __importStar(require("fs"));
const path = __importStar(require("path"));
const typescript_1 = __importDefault(require("typescript"));
const stubModuleSource = fs.readFileSync(path.join(__dirname, 'index.d.ts'), 'utf8');
function transformer(program) {
    return (context) => (source) => visitNodeAndChildren(source, program, context);
}
exports.default = transformer;
function visitNodeAndChildren(node, program, context) {
    const newNode = visitNode(node, program);
    return newNode
        ? typescript_1.default.visitEachChild(newNode, (child) => visitNodeAndChildren(child, program, context), context)
        : undefined;
}
function visitNode(node, program) {
    try {
        if (typescript_1.default.isCallExpression(node)) {
            return visitCallExpression(node, program);
        }
        if (typescript_1.default.isImportDeclaration(node)) {
            return visitImportClause(node, program);
        }
        return node;
    }
    catch (err) {
        if (err instanceof Error) {
            enhanceErrorStack(err, node);
        }
        throw err;
    }
}
function visitCallExpression(node, program) {
    const typeChecker = program.getTypeChecker();
    const signature = typeChecker.getResolvedSignature(node);
    if (!signature) {
        return node;
    }
    const { declaration } = signature;
    if (!declaration
        || typescript_1.default.isJSDocSignature(declaration)
        || !isOurStubModule(declaration.getSourceFile())) {
        return node;
    }
    const funcName = declaration.name && declaration.name.getText();
    if (!funcName) {
        return node;
    }
    return handleInlineCallExpression(node, funcName);
}
function visitImportClause(node, program) {
    if (!node.importClause) {
        return node;
    }
    const namedBindings = node.importClause.namedBindings;
    if (!node.importClause.name && !namedBindings) {
        return node;
    }
    const importSymbol = program.getTypeChecker().getSymbolAtLocation(node.moduleSpecifier);
    if (!importSymbol || !isOurStubModule(importSymbol.valueDeclaration.getSourceFile())) {
        return node;
    }
    return undefined; // drop the import
}
function handleInlineCallExpression(node, funcName) {
    const arg0 = node.arguments[0];
    if (!arg0 || !typescript_1.default.isStringLiteral(arg0)) {
        throw TypeError(`The first argument of ${funcName} function must be a string literal`);
    }
    const filename = arg0.text;
    const baseDir = path.dirname(node.getSourceFile().fileName);
    const content = fs.readFileSync(path.resolve(baseDir, filename), 'utf-8');
    switch (funcName) {
        case '$INLINE_FILE': {
            return typescript_1.default.createStringLiteral(content);
        }
        case '$INLINE_JSON': {
            const parent = node.parent;
            let obj = JSON.parse(content);
            if (typescript_1.default.isVariableDeclaration(parent) && typescript_1.default.isObjectBindingPattern(parent.name)) {
                if (typeof obj !== 'object') {
                    throw TypeError(`${filename} does not contain an object as expected`);
                }
                obj = filterObjectByBindingPattern(obj, parent.name);
            }
            return jsonToAST(obj);
        }
        default: {
            throw RangeError(`Unknown function: ${funcName}`);
        }
    }
}
function isOurStubModule(sourceFile) {
    // Comparing of the file names may not be reliable, it doesn't work e.g. with
    // yarn's "link" resolution used in this module's end-to-end tests.
    return sourceFile.text === stubModuleSource;
}
function filterObjectByBindingPattern(obj, binding) {
    return binding.elements.reduce((acc, { propertyName, name }) => {
        const propName = propertyName && typescript_1.default.isIdentifier(propertyName) ? propertyName.text
            : typescript_1.default.isIdentifier(name) ? name.text
                : undefined;
        if (propName) {
            acc[propName] = obj[propName];
        }
        return acc;
    }, {});
}
function jsonToAST(obj) {
    if (obj === null) {
        return typescript_1.default.createNull();
    }
    if (Array.isArray(obj)) {
        return typescript_1.default.createArrayLiteral(obj.map(jsonToAST));
    }
    if (typeof obj === 'object') {
        return typescript_1.default.createObjectLiteral(Object.keys(obj).map(key => {
            const propName = typescript_1.default.createStringLiteral(key);
            return typescript_1.default.createPropertyAssignment(propName, jsonToAST(obj[key]));
        }));
    }
    return typescript_1.default.createLiteral(obj);
}
function enhanceErrorStack(err, node) {
    if (err.stack == null) {
        return;
    }
    const lines = err.stack.split('\n');
    const line1 = lines[1] || '';
    const indent = ' '.repeat(line1.length - line1.trimLeft().length);
    const source = node.getSourceFile();
    const loc = typescript_1.default.getLineAndCharacterOfPosition(source, node.pos);
    const newLine = `${indent}at ${source.fileName}:${loc.line + 1}:${loc.character + 1}`;
    lines.splice(1, 0, newLine);
    err.stack = lines.join('\n');
}
//# sourceMappingURL=transformer.js.map